/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#include "ballonactivity.h"
#include <exception>
#include "exceptions.h"
#include <wx/msgdlg.h>
#include <wx/intl.h>
#include <SDL_gfxPrimitives.h>
#include "crvimage.h"
#include "ballonconfiguration.h"

#include <sys/timeb.h>
#include <sys/types.h>

#define MAX_SENSITIVY 800.0f
#define RAW_MOTION_NOISE_LEVEL 0.00001
#define RAW_MOTION_MAX_VALUE 0.03f
#define AUTOSENSITIVY_WAIT_TIME_MS 2000
//#define MOTION_THRESHOLD 0.0001f
#define MOTION_THRESHOLD 0.005f

CBallonActivity::CBallonActivity() : 
  m_patch(_T("ballon.pd"))
, m_pMhiTracker()
, m_oscOut(50000)
, m_smoothing(0.75f)
, m_baseFreq (220.0f)
, m_freqRange (1500.0f)
, m_backgroundColour(0)
, m_drawingColour(0xFFFFFFFF)
, m_sensitivy(MAX_SENSITIVY)
, m_autoSensitivy(true)
, m_motionThreshold(MOTION_THRESHOLD)
, m_motion_history_index(0)
, m_prevMotion(0)
, m_maxArea(50.0f)
, m_currentStage(CBallonActivity::BALLON_BW)
, m_invitationStage(0)
, m_invitationFired(false)
, m_invitationWaitTime(20000)
, m_invitationDuration(2000)
, m_firstTime(true)
, m_camProcessor(*this)
{		
	for (int i= 0; i< MOTION_HISTORY_SIZE; i++) m_motion_history[i]= 0;
//	m_pMhiTracker= new CMhiTracker();
}

CBallonActivity::~CBallonActivity(void)
{
//	CCameraController::getInstance()->UnregisterImageProcessor (this);

//	if (m_pMhiTracker) {
//		delete m_pMhiTracker;
//		m_pMhiTracker= NULL;
//	}

//	if (m_ballonConfiguration) delete m_ballonConfiguration;
}

void CBallonActivity::Run()
{
//	CCameraController::getInstance()->RegisterImageProcessor (this);
	m_camProcessor.Register();
}

void CBallonActivity::Finish()
{
//	CCameraController::getInstance()->UnregisterImageProcessor (this);
}

// Return timestamp in ms
_int64 CBallonActivity::GetTStamp (void) const
{
	struct timeb now;
	ftime(&now);
	return (_int64) now.time * 1000 + now.millitm;
}

// Adjustable parameters:
// m_smoothing
//
float CBallonActivity::ComputeRealMotion (CIplImage& image)
{
	float real_motion, motion;

	// 
	// Image processing
	//
	//m_pOfTracker->ProcessImage (&image, vx, vy);
	m_pMhiTracker.ProcessImage (&image, real_motion);

	//
	// Raw motion pre-processing
	//

	// Frame time normalization
	if (m_currentTStamp> m_lastTStamp)
		real_motion= real_motion / (float) (m_currentTStamp - m_lastTStamp);				
	else
		real_motion= 0;	
	
	//
	// Values range
	//

	if (real_motion< RAW_MOTION_NOISE_LEVEL) motion= 0;	// Noise level	
	else if (real_motion> RAW_MOTION_MAX_VALUE) real_motion= RAW_MOTION_MAX_VALUE;	// Filter values too high
	real_motion*= (1.0f / RAW_MOTION_MAX_VALUE);	// Normalize range

	
	//
	// Smoothing
	//	
	motion= m_prevMotion * m_smoothing + real_motion * (1.0f - m_smoothing);
	m_prevMotion= motion;

	//motion= log10(9.0f*motion + 1.0f);
	//if (motion< 0.1f) motion= 0.0f;
	//else if (motion> 1.0f) motion= 1.0f;

	return motion;
}

bool CBallonActivity::MotionDetected(float scaled_motion)
{
	return (scaled_motion> m_motionThreshold);	
}

// Returns motion * sensitivy after auto-regulating sensitivy
// when needed
float CBallonActivity::ComputeMotionXSensitivy (float real_motion)
{
	float scaled_motion;

	scaled_motion= real_motion * m_sensitivy;

	// TODO: considerar acotar entre 0 i 1
	if (!m_autoSensitivy)	
	{
		if (scaled_motion> 1.0f) scaled_motion= 1.0f;
		return scaled_motion;
	}

	//if (motion> 0.0001)
	//if (MotionDetected(real_motion))
	if (MotionDetected(scaled_motion))
	{
		m_motion_history[m_motion_history_index]= real_motion;
		m_motion_history_index= (m_motion_history_index + 1) % MOTION_HISTORY_SIZE;
		m_lastMotionTStamp= m_currentTStamp;	// Considerar canviar var
	}
	else
	{
		if (m_currentTStamp - m_lastMotionTStamp> AUTOSENSITIVY_WAIT_TIME_MS)
		{
			float mean_motion= 0;
			for (int i= 0; i< MOTION_HISTORY_SIZE; i++) mean_motion+= m_motion_history[i];
			mean_motion/= (float) MOTION_HISTORY_SIZE;

			if (mean_motion> 0)
				m_sensitivy = 0.9f / mean_motion;
			else
				m_sensitivy= MAX_SENSITIVY;

			if (m_sensitivy> MAX_SENSITIVY) m_sensitivy= MAX_SENSITIVY;

			// Sensitivy modified update scaled_motion
			scaled_motion= real_motion * m_sensitivy;
		}
	}

	// Reduce sensitivy if too high
	if (scaled_motion> 1.0f)
	{		
		m_sensitivy = 0.9f / real_motion;
		scaled_motion= real_motion * m_sensitivy;
	}		

	return scaled_motion;
}

/*
void CBallonActivity::PatchSendHarmony (float h)
{
	
	//if (area> 0.0f)
	//	freq= m_baseFreq + (m_freqRange * area / m_maxArea);
	//else
	//	freq= 0.0f;			
	
	*m_oscOut << osc::BeginBundleImmediate << osc::BeginMessage ("/harmony") 
		<< h << osc::EndMessage << osc::EndBundle;
	SendOSC();
	m_oscOut->Clear();
}

void CBallonActivity::PatchSendNumberHarmonics (float nh)
{
	*m_oscOut << osc::BeginBundleImmediate << osc::BeginMessage ("/number_harmonics") 
		<< nh << osc::EndMessage << osc::EndBundle;
	SendOSC();
	m_oscOut->Clear();
}*/

/*
void CBallonActivity::PatchSendMotion (float m)
{
	*m_oscOut << osc::BeginBundleImmediate << osc::BeginMessage ("/motion") 
		<< m << osc::EndMessage << osc::EndBundle;
	SendOSC();
	m_oscOut->Clear();
}

void CBallonActivity::PatchSendSize (float s)
{
	*m_oscOut << osc::BeginBundleImmediate << osc::BeginMessage ("/size") 
		<< s << osc::EndMessage << osc::EndBundle;
	SendOSC();
	m_oscOut->Clear();
}
*/

void CBallonActivity::PatchSend (float motion, float size)
{
	m_oscOut << osc::BeginBundleImmediate << osc::BeginMessage ("/ballon") 
		<< (int) m_currentStage << motion << size
		<< osc::EndMessage << osc::EndBundle;
	//SendOSC();
	//m_oscOut->Clear();
	m_oscOut.Send();
}

void CBallonActivity::ProcessImage (IplImage *pImage)
{
	static float real_motion= 0, scaled_motion= 0, area= 0;

	if (m_firstTime) {
		m_lastInvitationEndTStamp= m_invitationStartTStamp= m_currentTStamp= m_lastTStamp= m_lastMotionTStamp= GetTStamp ();
		m_firstTime= false;
	}

	CIplImage image(pImage);			
//	m_pCamDialog->DrawCam (pImage);

	real_motion= ComputeRealMotion (image);			
	scaled_motion= ComputeMotionXSensitivy (real_motion);
	if (MotionDetected(scaled_motion)) 
	{
		m_lastMotionTStamp= m_currentTStamp;
		m_lastInvitationEndTStamp= m_currentTStamp;
		m_invitationStage= 0;
		m_invitationFired= false;
	}
	else
	{
		if (!m_invitationFired)
		{
			if (m_currentTStamp - m_lastInvitationEndTStamp> m_invitationWaitTime)
			{
				// Wait time exceeded -> fire invitation
				m_invitationStage++;
				m_invitationFired= true;
				m_invitationStartTStamp= m_currentTStamp;
			}
		}
		else if (m_currentTStamp - m_invitationStartTStamp> m_invitationDuration)
		{
			// Invitation end
			m_lastInvitationEndTStamp= m_currentTStamp;
			m_invitationFired= false;
		}
	}

	if (m_invitationFired)
	{
		float size;
		float percent= (float) (m_currentTStamp - m_invitationStartTStamp) / (float) m_invitationDuration;

		if (percent< 0.25f)
		{
			// Raise
			size= percent * 4.0f;
		}
		else
		{
			// Fall
			size= 1.0f - (percent - 0.25f);
			size= size * 1.333333333333333f - 0.33333333f;						
		}

		//Draw (size, hsv2rgba (0, 1.0f, 1.0f), hsv2rgba (0.5f, 1.0f, 1.0f));
		if (m_sdlDrawer.GetSurface()) {
			m_simpleShapeDrawer.Draw(m_sdlDrawer.GetSurface(), size, hsv2rgba (0, 1.0f, 1.0f), hsv2rgba (0.5f, 1.0f, 1.0f));
		}
	}
	else
	{
		switch (m_currentStage)
		{
		case CBallonActivity::BALLON_BW:
			if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, scaled_motion); // PatchSendNumberHarmonics(scaled_motion);
			else PatchSend (0, 0); //PatchSendNumberHarmonics(0);
			if (m_sdlDrawer.GetSurface()) {
				m_simpleShapeDrawer.Draw(m_sdlDrawer.GetSurface(),scaled_motion, m_drawingColour, m_backgroundColour);
			}
			break;
		case CBallonActivity::BALLON_COLOR:
			{
			float h= 0.5f-scaled_motion;
			if (h< 0.0f) h+= 1.0f;
			if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, scaled_motion); //PatchSendNumberHarmonics(scaled_motion);
			else PatchSend (0, 0); //PatchSendNumberHarmonics(0);
			if (m_sdlDrawer.GetSurface()) {
				m_simpleShapeDrawer.Draw(m_sdlDrawer.GetSurface(),scaled_motion, hsv2rgba (scaled_motion, 1.0f, 1.0f), hsv2rgba (h, 1.0f, 1.0f));
			}
			}
			break;				
		case CBallonActivity::BALLON_BLOWUP:				
		case CBallonActivity::BALLON_BLOWUP_COLOR:
			area= area + scaled_motion;
			if (m_currentTStamp != m_lastMotionTStamp) 
			{
				float decay= m_maxArea * 0.002 * pow ((1.0f + 1.0f / 1400.0f), (float) (m_currentTStamp - m_lastMotionTStamp));
				area-= decay;
			}
			if (area< 0.0f) area= 0.0f;
			else if (area> m_maxArea) area= m_maxArea;				
//					PatchSendNumberHarmonics(area / m_maxArea);
			if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, area / m_maxArea); //PatchSendHarmony (scaled_motion);
			else PatchSend (0, area / m_maxArea);	//PatchSendHarmony (0);
			
			if (m_currentStage== CBallonActivity::BALLON_BLOWUP) {
				if (m_sdlDrawer.GetSurface()) {
					m_simpleShapeDrawer.Draw(m_sdlDrawer.GetSurface(),area / m_maxArea, m_drawingColour, m_backgroundColour);
				}
			}
			else
			{
				float h= 0.5f-scaled_motion;
				if (h< 0.0f) h+= 1.0f;
				if (m_sdlDrawer.GetSurface()) {
					m_simpleShapeDrawer.Draw(m_sdlDrawer.GetSurface(),area / m_maxArea, hsv2rgba (scaled_motion, 1.0f, 1.0f), hsv2rgba (h, 1.0f, 1.0f));
				}
			}
			break;
		
			
		default:
			throw RunTimeException(_("Ballon activity: no such stage"));
		}
	}
	if (m_sdlDrawer.GetSurface()) {
		m_sdlDrawer.FinishedDrawing();
	}

	m_lastTStamp= m_currentTStamp;		
	m_currentTStamp= GetTStamp ();
}

/*
void CBallonActivity::MainLoop(void)
{
	
	float real_motion= 0, scaled_motion= 0, area= 0; //, area_ant= 0;
	IplImage* pImage;	
	//m_pOfTracker= new COfTracker (m_pCamDialog->GetCamWindow());
	m_pMhiTracker= new CMhiTracker (m_pCamDialog->GetCamWindow());
	
    		
	m_lastInvitationEndTStamp= m_invitationStartTStamp= m_currentTStamp= m_lastTStamp= m_lastMotionTStamp= GetTStamp ();

	try{

		while (ContinueLoop())
		{
			// Image
			pImage= m_pCamera->QueryFrame();
			CIplImage image(pImage);			
			m_pCamDialog->DrawCam (pImage);

			real_motion= ComputeRealMotion (image);			
			scaled_motion= ComputeMotionXSensitivy (real_motion);
			if (MotionDetected(scaled_motion)) 
			{
				m_lastMotionTStamp= m_currentTStamp;
				m_lastInvitationEndTStamp= m_currentTStamp;
				m_invitationStage= 0;
				m_invitationFired= false;
			}
			else
			{
				if (!m_invitationFired)
				{
					if (m_currentTStamp - m_lastInvitationEndTStamp> m_invitationWaitTime)
					{
						// Wait time exceeded -> fire invitation
						m_invitationStage++;
						m_invitationFired= true;
						m_invitationStartTStamp= m_currentTStamp;
					}
				}
				else if (m_currentTStamp - m_invitationStartTStamp> m_invitationDuration)
				{
					// Invitation end
					m_lastInvitationEndTStamp= m_currentTStamp;
					m_invitationFired= false;
				}
			}

			if (m_invitationFired)
			{
				float size;
				float percent= (float) (m_currentTStamp - m_invitationStartTStamp) / (float) m_invitationDuration;

				if (percent< 0.25f)
				{
					// Raise
					size= percent * 4.0f;
				}
				else
				{
					// Fall
					size= 1.0f - (percent - 0.25f);
					size= size * 1.333333333333333f - 0.33333333f;						
				}

				//Draw (size, hsv2rgba (0, 1.0f, 1.0f), hsv2rgba (0.5f, 1.0f, 1.0f));
				m_simpleShapeDrawer.Draw(m_screen, size, hsv2rgba (0, 1.0f, 1.0f), hsv2rgba (0.5f, 1.0f, 1.0f));
			}
			else
			{
				switch (m_currentStage)
				{
				case CBallonActivity::BALLON_BW:
					if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, scaled_motion); // PatchSendNumberHarmonics(scaled_motion);
					else PatchSend (0, 0); //PatchSendNumberHarmonics(0);
					m_simpleShapeDrawer.Draw(m_screen,scaled_motion, m_drawingColour, m_backgroundColour);
					break;
				case CBallonActivity::BALLON_COLOR:
					{
					float h= 0.5f-scaled_motion;
					if (h< 0.0f) h+= 1.0f;
					if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, scaled_motion); //PatchSendNumberHarmonics(scaled_motion);
					else PatchSend (0, 0); //PatchSendNumberHarmonics(0);
					m_simpleShapeDrawer.Draw(m_screen,scaled_motion, hsv2rgba (scaled_motion, 1.0f, 1.0f), hsv2rgba (h, 1.0f, 1.0f));
					}
					break;				
				case CBallonActivity::BALLON_BLOWUP:				
				case CBallonActivity::BALLON_BLOWUP_COLOR:
					area= area + scaled_motion;
					if (m_currentTStamp != m_lastMotionTStamp) 
					{
						float decay= m_maxArea * 0.002 * pow ((1.0f + 1.0f / 1400.0f), (float) (m_currentTStamp - m_lastMotionTStamp));
						area-= decay;
					}
					if (area< 0.0f) area= 0.0f;
					else if (area> m_maxArea) area= m_maxArea;				
//					PatchSendNumberHarmonics(area / m_maxArea);
					if (MotionDetected(scaled_motion)) PatchSend (scaled_motion, area / m_maxArea); //PatchSendHarmony (scaled_motion);
					else PatchSend (0, area / m_maxArea);	//PatchSendHarmony (0);
					
					if (m_currentStage== CBallonActivity::BALLON_BLOWUP)
						m_simpleShapeDrawer.Draw(m_screen,area / m_maxArea, m_drawingColour, m_backgroundColour);
					else
					{
						float h= 0.5f-scaled_motion;
						if (h< 0.0f) h+= 1.0f;
						m_simpleShapeDrawer.Draw(m_screen,area / m_maxArea, hsv2rgba (scaled_motion, 1.0f, 1.0f), hsv2rgba (h, 1.0f, 1.0f));
					}
					break;
				
					
				default:
					throw RunTimeException(_("Ballon activity: no such stage"));
				}
			}
			m_lastTStamp= m_currentTStamp;		
			m_currentTStamp= GetTStamp ();			
		}
	}
	catch (...)
	{
		//delete m_pOfTracker;
		delete m_pMhiTracker;
		throw;
	}	
//	delete m_pOfTracker;
	delete m_pMhiTracker;
}
*/

//#define min(a,b) ((a)< (b)? (a) : (b))
//#define max(a,b) ((a)> (b)? (a) : (b))
#define min3(a,b,c) min(a, min(b,c))
#define max3(a,b,c) max(a, max(b,c))

void CBallonActivity::rgba2hsv (Uint32 rgba, float &h, float &s, float &v)
{
	float var_R = ( (rgba >> 24) / 255.0f );                     //RGB from 0 to 255
	float var_G = ( (rgba >> 16) / 255.0f );
	float var_B = ( (rgba >> 8) / 255.0f );

	float var_Min = min3( var_R, var_G, var_B );    //Min. value of RGB
	float var_Max = max3( var_R, var_G, var_B );    //Max. value of RGB
	float del_Max = var_Max - var_Min;             //Delta RGB value

	v = var_Max;

	if ( del_Max == 0 )                     //This is a gray, no chroma...
	{
		h= 0;
		s= 0;	
	}
	else                                    //Chromatic data...
	{
		s = del_Max / var_Max;

		float del_R = ( ( ( var_Max - var_R ) / 6.0f ) + ( del_Max / 2.0f ) ) / del_Max;
		float del_G = ( ( ( var_Max - var_G ) / 6.0f ) + ( del_Max / 2.0f ) ) / del_Max;
		float del_B = ( ( ( var_Max - var_B ) / 6.0f ) + ( del_Max / 2.0f ) ) / del_Max;

		if      ( var_R == var_Max ) h = del_B - del_G;
		else if ( var_G == var_Max ) h = ( 1.0f / 3.0f ) + del_R - del_B;
		else if ( var_B == var_Max ) h = ( 2.0f / 3.0f ) + del_G - del_R;

		if ( h < 0 ) h+= 1.0f;
		else if ( h > 1 ) h-= 1.0f;   
	}
}

Uint32 CBallonActivity::hsv2rgba (float h, float s, float v)
{
	float r, g, b;

	assert (h>= 0 && h<= 1.0f);
	assert (s>= 0 && s<= 1.0f);
	assert (v>= 0 && v<= 1.0f);

	if (s == 0 )                       //HSV from 0 to 1
	{
		r = v * 255.0f;
		g = v * 255.0f;
		b = v * 255.0f;
	}
	else
	{
		float var_h = h * 6.0f;

		if ( var_h== 6.0f ) var_h = 0;      //H must be < 1
		float var_i = floor ( var_h );             //Or ... var_i = floor( var_h )
		float var_1 = v * ( 1.0f - v );
		float var_2 = v * ( 1.0f - s * ( var_h - var_i ) );
		float var_3 = v * ( 1.0f - s * ( 1.0f - ( var_h - var_i ) ) );

		float var_r,  var_g, var_b;

		if      ( var_i == 0 ) { var_r = v    ; var_g = var_3; var_b = var_1; }
		else if ( var_i == 1 ) { var_r = var_2; var_g = v    ; var_b = var_1; }
		else if ( var_i == 2 ) { var_r = var_1; var_g = v    ; var_b = var_3; }
		else if ( var_i == 3 ) { var_r = var_1; var_g = var_2; var_b = v    ; }
		else if ( var_i == 4 ) { var_r = var_3; var_g = var_1; var_b = v    ; }
		else                   { var_r = v    ; var_g = var_1; var_b = var_2; }

		r = var_r * 255.0f;                  //RGB results from 0 to 255
		g = var_g * 255.0f;
		b = var_b * 255.0f;
	}
	assert (r<= 255.0f && g<= 255.0f && b<= 255.0f);

	Uint32 result= 0xFF;	// Alpha

	result= result | (((Uint32) r) << 24);
	result= result | (((Uint32) g) << 16);
	result= result | (((Uint32) b) << 8);
	
	return result;
}

void CBallonActivity::SetSmoothing (float smoothing)
{
	if (smoothing< 0.0f || smoothing> 1.0f) 
		throw RunTimeException(_("Smoothing out of range"));

	m_smoothing= log10(smoothing * 9.0f + 1.0f);
}

void CBallonActivity::SetBaseFreq (float freq)
{
	if (freq< 50.0f || freq> 10000.0f) 
		throw RunTimeException(_("Base freq out of range"));
	m_baseFreq= freq;
}

void CBallonActivity::SetFreqRange (float range)
{
	if (range< 50.0f || range> 10000.0f) 
		throw RunTimeException(_("Freq range out of range"));
	m_freqRange= range;
}

void CBallonActivity::SetBackgroundColour (unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
	//m_backgroundColour= SDL_MapRGBA(m_screen->format, r,g,b,a);	
	m_backgroundColour= r;
	m_backgroundColour<<= 8;
	m_backgroundColour+= g;
	m_backgroundColour<<= 8;
	m_backgroundColour+= b;
	m_backgroundColour<<= 8;
	m_backgroundColour+= a;	
}

void CBallonActivity::SetDrawingColour (unsigned char r, unsigned char g, unsigned char b, unsigned char a)
{
	//m_drawingColour= SDL_MapRGBA(m_screen->format, r,g,b,a);
	m_drawingColour= r;
	m_drawingColour<<= 8;
	m_drawingColour+= g;
	m_drawingColour<<= 8;
	m_drawingColour+= b;
	m_drawingColour<<= 8;
	m_drawingColour+= a;	
}
/*

void CBallonActivity::OpenConfigDialog(void)
{
	if (m_ballonConfiguration== NULL)
	{
		m_ballonConfiguration= new CBallonConfiguration (this, &m_simpleShapeDrawer);
		m_ballonConfiguration->Create (NULL);
	}
	m_ballonConfiguration->Show(true);
}*/
