/////////////////////////////////////////////////////////////////////////////
// Name:        capturethread2.cpp
// Purpose:		
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
#pragma implementation
#pragma interface
#endif

#ifndef  WX_PRECOMP
  #include "wx/wx.h"
#endif //precompiled headers

#include <cv.h>
#include <algorithm>
#include <exception>

#include "capturethread2.h"
#include "crvcamera.h"
#include "camwindow2.h"
#include "exceptions.h"

//
//
// Event to notify thread termination
//
//
IMPLEMENT_DYNAMIC_CLASS( myThreadEvent, wxEvent )

DEFINE_EVENT_TYPE( wxEVT_MY_THREAD_EVENT )

myThreadEvent::myThreadEvent( wxString msg, TFlags flags, CProcessImage2* who)
: m_msg(msg)
, m_flags(flags)
, m_who(who) 
{ 
	SetEventType ( wxEVT_MY_THREAD_EVENT );
	SetEventObject (this);
}
//
// End event stuff
//

// Constructor
CCaptureThread2::CCaptureThread2 (CCamera *pCamera, wxThreadKind kind, wxEvtHandler* evtDest)
: m_pCamera (pCamera)
, wxThread (kind)
, m_eventDest(evtDest)
, m_Life (true)
{
	assert (m_pCamera!= NULL);
}

// Destructor
CCaptureThread2::~CCaptureThread2 ()
{
}

int CCaptureThread2::RegisterImageProcessor (CProcessImage2* iProcessor)
{
	wxCriticalSectionLocker mutex(m_mutex);
	return DoRegisterImageProcessor (iProcessor);	
}

int CCaptureThread2::UnregisterImageProcessor (CProcessImage2* iProcessor)
{
	wxCriticalSectionLocker mutex(m_mutex);
	return DoUnregisterImageProcessor (iProcessor);
}

int CCaptureThread2::GetRegisteredImageProcessors ()
{
	wxCriticalSectionLocker mutex(m_mutex);
	return m_ImageProcessorList.size();
}

int CCaptureThread2::RegisterImageDrawer (CCamWindow2* iDrawer)
{
	wxCriticalSectionLocker mutex(m_mutex);
	return DoRegisterImageDrawer (iDrawer);
}

int CCaptureThread2::UnregisterImageDrawer (CCamWindow2* iDrawer)
{
	wxCriticalSectionLocker mutex(m_mutex);
	return DoUnregisterImageDrawer (iDrawer);
}

int CCaptureThread2::GetRegisteredImageDrawers ()
{
	wxCriticalSectionLocker mutex(m_mutex);
	return m_ImageDrawerList.size();
}

int CCaptureThread2::DoRegisterImageProcessor (CProcessImage2* iProcessor)
{
	assert (iProcessor);

	TImageProcessorListIterator it;

	// Checks if already exists
	it= find (m_ImageProcessorList.begin(), m_ImageProcessorList.end(), iProcessor);
	
	// Adds
	if (it== m_ImageProcessorList.end())
		m_ImageProcessorList.push_back (iProcessor);

	return m_ImageProcessorList.size();
}

int CCaptureThread2::DoUnregisterImageProcessor (CProcessImage2* iProcessor)
{
	assert (iProcessor);
	TImageProcessorListIterator it= find (m_ImageProcessorList.begin(), m_ImageProcessorList.end(), iProcessor);
	if (it!= m_ImageProcessorList.end()) m_ImageProcessorList.erase (it);
	return m_ImageProcessorList.size();
}

int CCaptureThread2::DoRegisterImageDrawer (CCamWindow2* iDrawer)
{
	assert (iDrawer);
	TImageDrawerListIterator it= find (m_ImageDrawerList.begin(), m_ImageDrawerList.end(), iDrawer);
	if (it== m_ImageDrawerList.end()) m_ImageDrawerList.push_back (iDrawer);
	return m_ImageDrawerList.size();
}

int CCaptureThread2::DoUnregisterImageDrawer (CCamWindow2* iDrawer)
{
	assert (iDrawer);
	TImageDrawerListIterator it= find (m_ImageDrawerList.begin(), m_ImageDrawerList.end(), iDrawer);
	if (it!= m_ImageDrawerList.end()) m_ImageDrawerList.erase (it);
	return m_ImageDrawerList.size();
}

wxThread::ExitCode CCaptureThread2::Entry( )
{
	wxThread::ExitCode exitCode= NULL;

	if (m_eventDest) {
			myThreadEvent event (_T(""), (myThreadEvent::TFlags)(myThreadEvent::STARTED));
			wxPostEvent(m_eventDest, event);
	}

	// Camera shold be opened inside calling thread
	if (!m_pCamera->Open())
	{
		// This code never should be reached, but it may happen: there is a race condition between
		// testing camera availabiliy and effectively using it.
		if (m_eventDest) {
			myThreadEvent event (_("Camera not detected."), 
				(myThreadEvent::TFlags) (myThreadEvent::ENDED | myThreadEvent::FATAL_ERROR) );
			wxPostEvent(m_eventDest, event);
		}
		return (wxThread::ExitCode) -1;
	}

	// Caught exceptions are reported to the evtHanler. This list is used
	// to store them until the critical section is exited. There are two
	// main reasons to do this (i) to exit the critical section as soon as
	// possible and (ii) to avoid deadlocks when calling wxPostEvent (see
	// camwindow.cpp for further details).
	std::list<myThreadEvent*> pendingEvents;

	// Start thread main loop
	while (m_Life && !TestDestroy())
	{
		IplImage* pImg= m_pCamera->QueryFrame();
		if (pImg== NULL) {
			m_Life= false;
			if (m_eventDest) {
				myThreadEvent event (_("Error grabbing from camera."), 
					(myThreadEvent::TFlags) (myThreadEvent::ENDED | myThreadEvent::FATAL_ERROR) );
				wxPostEvent(m_eventDest, event);
			}
			exitCode= (wxThread::ExitCode) -1;
			break;
		}

		m_mutex.Enter();

		TImageProcessorListIterator it;
		for(it= m_ImageProcessorList.begin(); it != m_ImageProcessorList.end(); ++it) {
			try {		
				(*it)->ProcessImage (pImg);
			}
			catch (BaseException &e) {
				// Add pending event
				if (m_eventDest) {
					myThreadEvent* event= new myThreadEvent (e.getMessage(), 
						(myThreadEvent::TFlags) (myThreadEvent::NON_FATAL_ERROR), (*it));
					pendingEvents.push_back(event);
				}

				// Remove subscriber from list
				m_ImageProcessorList.erase(it);
			}
			catch (exception &e) {
				// Add pending event
				if (m_eventDest) {
					myThreadEvent* event= new myThreadEvent (wxString(e.what(), wxConvLocal),
						(myThreadEvent::TFlags) (myThreadEvent::NON_FATAL_ERROR), (*it));
					pendingEvents.push_back(event);
				}

				// Remove subscriber from list
				m_ImageProcessorList.erase(it);
			}
		}

		// Call drawers. TODO: inside Draw wxPostEvent is called so it might
		// cause a deadlock on some platforms (e.g. gtk1). We assume that
		// drawer will not report exceptions.
		TImageDrawerListIterator dit;
		for(dit= m_ImageDrawerList.begin(); dit != m_ImageDrawerList.end(); ++dit)
			(*dit)->DrawCam (pImg);

		m_mutex.Leave();

		// Send pending events		
		if (m_eventDest) {
			std::list<myThreadEvent*>::iterator pei;
			for (pei= pendingEvents.begin(); pei!= pendingEvents.end(); ++pei) {
				wxPostEvent(m_eventDest, *(*pei));
				delete (*pei);
			}
			pendingEvents.clear();
		}		
	}

	m_pCamera->Close();

	if (m_eventDest) {
		myThreadEvent event (_T(""), (myThreadEvent::TFlags)(myThreadEvent::ENDED));
		wxPostEvent(m_eventDest, event);
	}

	return exitCode;
}
